import json
import numpy as np
from uuid import uuid4
from pathlib import Path
from scipy.stats import chi2
import dash_html_components as html
from webviz_config.containers import WebvizContainer
import webviz_subsurface_components as wsc
from ..datainput import extract_mismatch


class HistoryMatch(WebvizContainer):
    '''### History match

This container visualizes the quality of the history match.

* `ensembles`: List of the ensembles in `container_settings` to visualize.
* `observation_File`: Path to the observation `.yaml` file.
* `title`: Optional title for the container.
'''

    def __init__(self, container_settings, ensembles, observation_file: Path):

        self.observation_file = observation_file
        self.ens_paths = tuple((ens,
                                container_settings['scratch_ensembles'][ens])
                               for ens in ensembles)

        data = extract_mismatch(self.ens_paths, self.observation_file)
        self.hm_data = json.dumps(self._prepareData(data))

        self.hm_id = 'hm-id-{}'.format(uuid4())

    def add_webvizstore(self):
        return [(extract_mismatch, [{'ens_paths': self.ens_paths,
                                     'observation_file': self.observation_file
                                     }])]

    def _prepareData(self, data):
        data = data.copy().reset_index()

        ensemble_labels = data.ensemble_name.unique().tolist()
        num_obs_groups = len(data.obs_group_name.unique())

        data['avg_pos'] = data['total_pos'] / data['number_data_points']
        data['avg_neg'] = data['total_neg'] / data['number_data_points']

        iterations = []
        for ensemble in ensemble_labels:
            df = data[data.ensemble_name == ensemble]
            iterations.append(df.groupby('obs_group_name').mean())

        sorted_iterations = HistoryMatch._sortIterations(iterations)

        iterations_dict = HistoryMatch._iterations_to_dict(sorted_iterations,
                                                           ensemble_labels)

        confidence_sorted = _get_sorted_edges(num_obs_groups)
        confidence_unsorted = _get_unsorted_edges()

        data = {}
        data['iterations'] = iterations_dict
        data['confidence_interval_sorted'] = confidence_sorted
        data['confidence_interval_unsorted'] = confidence_unsorted

        return data

    @staticmethod
    def _sortIterations(iterations):
        sorted_data = []

        for df in iterations:
            sorted_df = df.copy()

            sorted_data.append(
                sorted_df.assign(f=sorted_df['avg_pos'] + sorted_df['avg_neg'])
                         .sort_values('f', ascending=False)
                         .drop('f', axis=1)
            )

        return sorted_data

    @staticmethod
    def _iterations_to_dict(iterations, labels):
        retval = []

        for iteration, label in zip(iterations, labels):
            retval.append({
                'name': label,
                'positive': iteration['avg_pos'].tolist(),
                'negative': iteration['avg_neg'].tolist(),
                'labels': iteration.index.tolist()
            })

        return retval

    @property
    def layout(self):
        return html.Div([
                  wsc.HistoryMatch(id=self.hm_id, data=self.hm_data)
               ])


def _get_unsorted_edges():
    """P10 - P90 unsorted edge coordinates"""

    retval = {
        'low': chi2.ppf(0.1, 1),
        'high': chi2.ppf(0.9, 1)
    }

    return retval


def _get_sorted_edges(number_observation_groups):
    """P10 - P90 sorted edge coordinates"""

    monte_carlo_iterations = 100000

    sorted_values = np.empty((number_observation_groups,
                              monte_carlo_iterations))

    for i in range(monte_carlo_iterations):
        sorted_values[:, i] = np.sort(np.random.chisquare(
                                      df=1,
                                      size=number_observation_groups))

    sorted_values = np.flip(sorted_values, 0)

    P10 = np.percentile(sorted_values, 90, axis=1)
    P90 = np.percentile(sorted_values, 10, axis=1)

    # Dictionary with two arrays (P10, P90). Each array of length equal
    # to number of observation groups i.e. number of items along y axis.
    # These values are to be used for drawing the stair stepped
    # sorted P10-P90 area:

    coordinates = {'low': list(P10), 'high': list(P90)}

    return coordinates



from uuid import uuid4
from pathlib import Path
import pandas as pd
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output
from webviz_config.webviz_store import webvizstore
from webviz_config.common_cache import cache
from webviz_config.containers import WebvizContainer
from webviz_subsurface_components import Morris


class MorrisPlot(WebvizContainer):
    '''### Morris

This container renders a visualization of the Morris sampling method.
The Morris method can be used to screen parameters for how they
influence model response, both individually and through interaction
effect with other parameters.

* `csv_file`: Input data
* `title`: Optional title for the container.
'''

    def __init__(self, app, csv_file: Path):

        self.graph_id = 'graph-{}'.format(uuid4())
        self.vector_id = 'vector-{}'.format(uuid4())
        self.csv_file = csv_file
        self.data = read_csv(self.csv_file)
        self.vector_names = self.data['name'].unique()
        self.set_callbacks(app)

    @property
    def layout(self):
        return html.Div([
            html.Label('Vector', style={'font-size': '2rem'}),
            dcc.Dropdown(id=self.vector_id,
                         clearable=False,
                         options=[{'label': i, 'value': i}
                                  for i in list(self.vector_names)],
                         value=self.vector_names[0]),
            Morris(
                id=self.graph_id)
        ])

    def add_webvizstore(self):
        return [(read_csv, [{'csv_file': self.csv_file}])]

    def set_callbacks(self, app):
        @app.callback([
            Output(self.graph_id, 'output'),
            Output(self.graph_id, 'parameter'),
            Output(self.graph_id, 'parameters')],
            [Input(self.vector_id, 'value')])
        def update_plot(vector):
            df = self.data[self.data['name'] == vector]
            df = df.sort_values('time')
            output = df[['mean', 'max', 'min', 'time']]\
                .drop_duplicates()\
                .to_dict(orient='records')
            parameters = []

            for name in self.data['name'].unique():
                if name != vector:
                    name_df = self.data[self.data['name'] == name]
                    parameters.append({
                        'main': list(name_df['morris_main']),
                        'name': str(name),
                        'interactions': list(name_df['morris_interaction'])
                    })
            return output, vector, parameters


@cache.memoize(timeout=cache.TIMEOUT)
@webvizstore
def read_csv(csv_file) -> pd.DataFrame:
    return pd.read_csv(csv_file)



try:
    import fmu.ensemble
except ImportError:  # fmu.ensemble is an optional dependency, e.g.
    pass             # for a portable webviz instance, it is never used.

import pandas as pd
from pathlib import Path
from webviz_config.common_cache import cache
from webviz_config.webviz_store import webvizstore


@cache.memoize(timeout=cache.TIMEOUT)
def scratch_ensemble(ensemble_name, ensemble_path):
    return fmu.ensemble.ScratchEnsemble(ensemble_name, ensemble_path)


@cache.memoize(timeout=cache.TIMEOUT)
@webvizstore
def extract_mismatch(ens_paths, observation_file: Path) -> pd.DataFrame:
    """Convert the fmu-ensemble mismatch dataframe into the the format
    suitable for the interactive history match visualization.
    """

    list_ens = [scratch_ensemble(ensemble_name, path)
                for (ensemble_name, path) in ens_paths]

    ens_data = fmu.ensemble.EnsembleSet("HistoryMatch", list_ens)

    df_mismatch = fmu.ensemble.Observations(str(observation_file))\
                     .mismatch(ens_data)

    df_mismatch['NORMALISED_MISMATCH'] = \
        df_mismatch['L2'] / (df_mismatch['MEASERROR'] ** 2)

    # Create a dataframe containing number of
    # observation points within each observation key:
    df_count = df_mismatch.groupby(['OBSKEY', 'REAL', 'ENSEMBLE'])\
                          .size()\
                          .to_frame('COUNT')\
                          .reset_index()\
                          .drop_duplicates(['OBSKEY'], keep='first')\
                          .drop(columns=['REAL', 'ENSEMBLE'])

    # 1) Sum the normalised misfit (grouped by obskey, misfit sign
    #    realizaton and ensemble.
    # 2) Pivot the dataframe such that instead of two rows wrt. positive and
    #    negative misfit, we get two columns.
    # 3) Replace NaN values with 0 (NaN happens e.g. for the summed negative
    #    misfit if e.g. all misfit values are positive.
    # 4) Drop the column name 0 (webviz don't need summed misfit over all
    #    observation points with zero misfit :p)
    # 5) Merge in the COUNT column.
    # 6) Rename columns such that the columns from fmu.ensemble corresponds
    #    to those used in the webviz history match visualization.
    return df_mismatch.groupby(['OBSKEY', 'SIGN', 'REAL', 'ENSEMBLE'])\
                      .sum()[['NORMALISED_MISMATCH']]\
                      .pivot_table(index=['OBSKEY', 'REAL', 'ENSEMBLE'],
                                   columns='SIGN',
                                   values='NORMALISED_MISMATCH'
                                   )\
                      .reset_index()\
                      .fillna(0)\
                      .drop(columns=[0], errors='ignore')\
                      .merge(df_count, on='OBSKEY', how='left')\
                      .rename(columns={'OBSKEY': 'obs_group_name',
                                       'REAL': 'realization',
                                       'ENSEMBLE': 'ensemble_name',
                                       'COUNT': 'number_data_points',
                                       1: 'total_pos',
                                       -1: 'total_neg'})



import pandas as pd
from webviz_config.common_cache import cache
from webviz_config.webviz_store import webvizstore
from ..datainput import scratch_ensemble


@cache.memoize(timeout=cache.TIMEOUT)
@webvizstore
def get_summary_data(ensemble_paths: tuple, sampling: str,
                     column_keys: tuple) -> pd.DataFrame:
    """ Loops over given ensemble paths, extracts smry-data and concates them
    into one big df. An additional column ENSEMBLE is added for each
    ens-path to seperate the ensambles.
    column_keys is converted to list as list-type is needed in
    .get_smry_stats()"""

    column_keys = list(column_keys) if isinstance(
        column_keys, (list, tuple)) else None

    smry_data = []
    for ens, ens_path in ensemble_paths:
        ens_smry_data = scratch_ensemble(
            ens, ens_path).get_smry(
                time_index=sampling, column_keys=column_keys)
        ens_smry_data['ENSEMBLE'] = ens
        smry_data.append(ens_smry_data)
    return pd.concat(smry_data)


@cache.memoize(timeout=cache.TIMEOUT)
@webvizstore
def get_summary_stats(ensemble_paths: tuple, sampling: str,
                      column_keys: tuple) -> pd.DataFrame:
    """ Loops over given ensemble paths, extracts smry-data and concates them
    into one big df. An additional column ENSEMBLE is added for each
    ens-path to seperate the ensambles.
    column_keys is converted to list as list-type is needed in
    .get_smry_stats()"""

    column_keys = list(column_keys) if isinstance(
        column_keys, (list, tuple)) else None

    smry_stats = []
    for ens, ens_path in ensemble_paths:
        ens_smry_stats = scratch_ensemble(
            ens, ens_path).get_smry_stats(
                time_index=sampling, column_keys=column_keys)
        ens_smry_stats['ENSEMBLE'] = ens
        smry_stats.append(ens_smry_stats)

    return pd.concat(smry_stats)



import mock
import dash
import pandas as pd
from webviz_config.common_cache import cache
from webviz_config.containers import ParameterDistribution

# mocked functions
get_parameters = 'webviz_subsurface.containers'\
                '._parameter_distribution.get_parameters'


def test_parameter_dist(dash_duo):

    app = dash.Dash(__name__)
    app.css.config.serve_locally = True
    app.scripts.config.serve_locally = True
    app.config.suppress_callback_exceptions = True
    cache.init_app(app.server)
    container_settings = {'scratch_ensembles': {'iter-0': ''}}
    ensembles = ['iter-0']

    with mock.patch(get_parameters) as mock_parameters:
        mock_parameters.return_value = pd.read_csv('tests/data/parameters.csv')

        p = ParameterDistribution(app, container_settings, ensembles)

        app.layout = p.layout
        dash_duo.start_server(app)

        my_component = dash_duo.find_element(f'#{p.ens_matrix_id}')

        if not my_component.text.startswith('iter-0'):
            raise AssertionError()



from uuid import uuid4
import numpy as np
import dash_html_components as html
import dash_core_components as dcc
import dash_table
from dash.dependencies import Input, Output
from webviz_config.common_cache import cache
from webviz_config.containers import WebvizContainer
from ..datainput import extract_volumes


class InplaceVolumes(WebvizContainer):
    '''### Volumetrics

This container visualizes inplace volumetrics results from
csv files stored on standard format.

* `ensembles`: Which ensembles in `container_settings` to visualize.
* `volfiles`:  Key/value pair of csv files E.g. (geogrid: geogrid--oil.csv)
* `volfolder`: Optional local folder for csv files
* `response`: Optional initial visualized volume response

'''

    RESPONSES = {
        'BULK_OIL': 'Bulk Volume (Oil)',
        'NET_OIL': 'Net Volume (Oil)',
        'PORE_OIL': 'Pore Volume (Oil)',
        'HCPV_OIL': 'Hydro Carbon Pore Volume (Oil)',
        'STOIIP_OIL': 'Stock Tank Oil Initially Inplace',
        'BULK_GAS': 'Bulk Volume (Gas)',
        'NET_GAS': 'Net Volume (Gas)',
        'PORE_GAS': 'Pore Volume (Gas)',
        'HCPV_GAS': 'Hydro Carbon Pore Volume (Gas)',
        'GIIP_GAS': 'Gas Initially in-place',
        'RECOVERABLE_OIL': 'Recoverable Volume (Oil)',
        'RECOVERABLE_GAS': 'Recoverable Volume (Gas)'
    }

    def __init__(self, app, container_settings, ensembles: list,
                 volfiles: dict, volfolder: str = 'share/results/volumes',
                 response: str = 'STOIIP_OIL'):

        self.ens_paths = tuple((ens,
                                container_settings['scratch_ensembles'][ens])
                               for ens in ensembles)
        self.volfiles = tuple(volfiles.items())
        self.volfolder = volfolder
        self.initial_response = response
        self.volumes = extract_volumes(self.ens_paths,
                                       self.volfolder,
                                       self.volfiles)
        self.radio_plot_type_id = 'radio-plot-type-{}'.format(uuid4())
        self.response_id = 'response-{}'.format(uuid4())
        self.chart_id = 'chart-{}'.format(uuid4())
        self.radio_selectors_id = 'radio-selectors-{}'.format(uuid4())
        self.selectors_id = {
            x: str(uuid4())
            for x in self.selectors}
        self.table_cols = ['response', 'selector', 'mean',
                           'stddev', 'minimum', 'p90', 'p10', 'maximum']

        self.set_callbacks(app)

    def add_webvizstore(self):
        return [(extract_volumes, [
            {
                'ensemble_paths': self.ens_paths,
                'volfolder': self.volfolder,
                'volfiles': self.volfiles
            }])]

    @property
    def vol_columns(self):
        '''List of all columns in dataframe'''
        return list(self.volumes.columns)

    @property
    def all_selectors(self):
        '''List of all possible selectors'''
        return ['SOURCE', 'ENSEMBLE', 'ZONE', 'REGION', 'FACIES', 'LICENSE']

    @property
    def plot_types(self):
        '''List of available plots'''
        return ['Histogram', 'Per realization', 'Box Plot', 'Table']

    @property
    def selectors(self):
        '''List of available selector columns in dframe'''
        return [x
                for x in self.all_selectors
                if x in self.vol_columns]

    @property
    def responses(self):
        '''List of available volume responses in dframe'''
        return [x
                for x in self.vol_columns
                if x not in self.selectors and x != 'REAL']

    @property
    def vol_callback_inputs(self):
        '''Returns all Dash inputs for selecting and filtering volumes
        The number of inputs will vary depending on the available
        selector columns in the volumes dataframe
        '''
        inputs = []
        inputs.append(Input(self.response_id, 'value'))
        inputs.append(Input(self.radio_plot_type_id, 'value'))
        inputs.append(Input(self.radio_selectors_id, 'value'))
        for selector in self.selectors:
            inputs.append(Input(
                self.selectors_id[selector], 'value'))
        return inputs

    @property
    def selector_dropdowns(self):
        '''Makes dropdowns for each selector.
        Args:
            dframe - Volumetrics Dataframe
            selectors - List of selector columns
        Return:
            dcc.Dropdown objects
        '''
        dropdowns = []
        for selector in self.selectors:
            elements = list(self.volumes[selector].unique())
            multi = True

            value = elements
            if selector == 'ENSEMBLE' or selector == 'SOURCE':
                value = elements[0]
            dropdowns.append(
                html.Div(children=[
                    html.Details(open=True, children=[
                        html.Summary(selector.lower().capitalize()),
                        dcc.Dropdown(
                            id=self.selectors_id[selector],
                            options=[{'label': i, 'value': i}
                                     for i in elements],
                            value=value,
                            multi=multi)
                    ])
                ])
            )
        return dropdowns

    @property
    def style_plot_options(self):
        '''Simple grid layout for the selector row'''
        return {
            'display': 'grid',
            'align-content': 'space-around',
            'justify-content': 'space-between',
            'grid-template-columns': '2fr 1fr 1fr 1fr'
        }

    @property
    def style_layout(self):
        '''Simple grid layout for the main elements'''
        return {
            'display': 'grid',
            'align-content': 'space-around',
            'justify-content': 'space-between',
            'grid-template-columns': '5fr 1fr'
        }

    @staticmethod
    def group_radio_options(selectors):
        '''Returns options for a radio button used for grouping
        volume results on a dataframe column.
        The option 'NONE' is added to allow no grouping
        '''
        options = ['NONE']
        options.extend(selectors)
        return [{'label': i.lower().capitalize(), 'value': i} for i in options]

    @property
    def plot_options_layout(self):
        '''Row layout of dropdowns for plot options'''
        return html.Div(style=self.style_plot_options, children=[
                        html.Div(children=[
                            html.P('Response:', style={
                                'font-weight': 'bold'}),
                            dcc.Dropdown(
                                id=self.response_id,
                                options=[
                                    {'label': InplaceVolumes.RESPONSES.get(
                                        i, i), 'value': i}
                                    for i in self.responses],
                                value=self.initial_response
                                if self.initial_response in self.responses
                                else self.responses[0])
                        ]),
                        html.Div(children=[
                            html.P('Plot type:', style={
                                'font-weight': 'bold'}),
                            dcc.Dropdown(
                                id=self.radio_plot_type_id,
                                options=[{'label': i, 'value': i}
                                         for i in self.plot_types],
                                value='Per realization'),
                        ]),
                        html.Div(children=[
                            html.P('Group by:', style={
                                'font-weight': 'bold'}),
                            dcc.Dropdown(
                                id=self.radio_selectors_id,
                                options=self.group_radio_options(
                                    self.selectors),
                                value='NONE')]),
                        ])

    @property
    def layout(self):
        '''Main layout'''
        return html.Div([
            html.Div(style=self.style_layout, children=[
                html.Div(children=[
                    self.plot_options_layout,
                    html.Div(id=self.chart_id),
                ]),
                html.Div(children=[
                    html.P('Filters:', style={'font-weight': 'bold'}),
                    html.Div(
                        children=self.selector_dropdowns
                    ),
                ])

            ]),
        ])

    def set_callbacks(self, app):
        @app.callback(
            Output(self.chart_id, 'children'),
            self.vol_callback_inputs)
        def render_vol_chart(*args):
            '''Renders a volume visualization which could either by a dcc.Graph
            or a Dash table object.
            The arguments are given by the vol_callback_inputs property
            Args:
                response: The volumetrics response to plot
                plot_type: The type of graph/table to plot
                group: The selector to group the data by
                selections: Active values from the selector columns
            Return:
                dcc.Graph/dash_table.DataTable
            '''
            response = args[0]
            plot_type = args[1]
            group = args[2]
            selections = args[3:]
            data = self.volumes
            data = filter_dataframe(data, self.selectors, selections)
            # If not grouped make one trace
            if group == 'NONE':
                dframe = data.groupby('REAL').sum().reset_index()
                traces = [plot_data(plot_type, dframe, response, 'Total')]

            # Else make one trace for each group member
            else:
                traces = []
                for name, vol_group_df in data.groupby(group):
                    dframe = vol_group_df.groupby('REAL').sum().reset_index()
                    trace = plot_data(plot_type, dframe, response, name)
                    if trace is not None:
                        traces.append(trace)
            # Make a dash table if table is selected
            if plot_type == 'Table':
                return dash_table.DataTable(
                    columns=[{'name': i, 'id': i} for i in self.table_cols],
                    data=traces)
            # Else make a graph object
            else:
                return dcc.Graph(figure={
                    'data': traces,
                    'layout': plot_layout(plot_type, response)}
                )

        @app.callback(
            [Output(self.selectors_id['ENSEMBLE'], 'multi'),
             Output(self.selectors_id['ENSEMBLE'], 'value')],
            [Input(self.radio_selectors_id, 'value')])
        def set_iteration_selector(group_by):
            '''If iteration is selected as group by set the iteration
            selector to allow multiple selections, else use single selection
            '''
            if group_by == 'ENSEMBLE':
                return True, list(self.volumes['ENSEMBLE'].unique())
            else:
                return False, list(self.volumes['ENSEMBLE'].unique())[0]

        @app.callback(
            [Output(self.selectors_id['SOURCE'], 'multi'),
             Output(self.selectors_id['SOURCE'], 'value')],
            [Input(self.radio_selectors_id, 'value')])
        def set_source_selector(group_by):
            '''If iteration is selected as group by set the iteration
            selector to allow multiple selections, else use single selection
            '''
            if group_by == 'SOURCE':
                return True, list(self.volumes['SOURCE'].unique())
            else:
                return False, list(self.volumes['SOURCE'].unique())[0]


@cache.memoize(timeout=cache.TIMEOUT)
def plot_data(plot_type, dframe, response, name):
    values = dframe[response]

    if plot_type == 'Histogram':
        if values.nunique() == 1:
            values = values[0]
        return {
            'x': values,
            'type': 'histogram',
            'name': name
        }
    if plot_type == 'Box Plot':
        return {
            'y': values,
            'name': name,
            'type': 'box'
        }
    if plot_type == 'Per realization':
        return {
            'y': values,
            'x': dframe['REAL'],
            'name': name,
            'type': 'bar'
        }
    values.replace(0, np.nan, inplace=True)
    if plot_type == 'Table':
        try:
            return {
                'response': response,
                'selector': str(name),
                'minimum': f'{values.min():.2e}',
                'maximum': f'{values.max():.2e}',
                'mean': f'{values.mean():.2e}',
                'stddev': f'{values.std():.2e}',
                'p10': f'{np.percentile(values, 90):.2e}',
                'p90': f'{np.percentile(values, 10):.2e}'
            }
        except KeyError:
            return None


@cache.memoize(timeout=cache.TIMEOUT)
def plot_layout(plot_type, response):

    if plot_type == 'Histogram':
        return {
            'bargap': 0.05,
            'xaxis': {'title': response},
            'yaxis': {'title': 'Count'}
        }
    if plot_type == 'Box Plot':
        return {'yaxis': {'title': response}}

    if plot_type == 'Per Realization':
        return {
            'margin': {
                'l': 40,
                'r': 40,
                'b': 30,
                't': 10},
            'yaxis': {'title': response},
            'xaxis': {'title': 'Realization'}}


@cache.memoize(timeout=cache.TIMEOUT)
def filter_dataframe(dframe, columns, column_values):
    df = dframe.copy()
    if not isinstance(columns, list):
        columns = [columns]
    for filt, col in zip(column_values, columns):
        if isinstance(filt, list):
            df = df.loc[df[col].isin(filt)]
        else:
            df = df.loc[df[col] == filt]
    return df



from uuid import uuid4
from glob import glob
from pathlib import PurePath
from collections import OrderedDict
import numpy as np
import pandas as pd
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from dash_table import DataTable
from webviz_config.common_cache import cache
from webviz_config.containers import WebvizContainer
from ..datainput import scratch_ensemble, load_surface, get_wfence, get_hfence


class Intersect(WebvizContainer):
    '''### Intersect

This container visualizes surfaces intersected along a well path.
The input are surfaces from a FMU ensemble stored on standardized
format with standardized naming (share/results/maps/name--category.gri)
and a folder of well files stored in RMS well format.

* `ensemble`: Which ensemble in `container_settings` to visualize.
* `well_path`: File folder with wells in rmswell format
* `surface_cat`: Surface category to look for in the file names
* `surface_names`: List of surface names to look for in the file names
* `well_suffix`:  Optional suffix for well files. Default is .rmswell
'''
    COLORS = ['#543005', '#8c510a', '#bf812d', '#dfc27d',
              '#f6e8c3', '#f5f5f5', '#c7eae5', '#80cdc1',
              '#35978f', '#01665e', '#003c30']

    LAYOUT_STYLE = {
        'display': 'grid',
        'align-content': 'space-around',
        'justify-content': 'space-between',
        'grid-template-columns': '2fr 6fr',
    }

    FENCE_OPTION_STYLE = {
        'display': 'grid',
        'align-content': 'space-around',
        'justify-content': 'space-between',
        'grid-template-columns': '1fr 1fr',
        'max-width': '50%'
    }

    TABLE_STYLE = {
        'maxHeight': '300',
        'overflowY': 'auto'
    }

    def __init__(self, app, container_settings, ensemble,
                 well_path, surface_cat, surface_names,
                 well_suffix='.rmswell'):

        self.well_path = well_path
        self.ensemble_path = container_settings['scratch_ensembles'][ensemble]
        self.ensemble = scratch_ensemble(ensemble, self.ensemble_path)
        self.well_suffix = well_suffix
        self.surface_cat = surface_cat
        self.surface_names = surface_names
        self.unique_id = f'{uuid4()}'
        self.well_list_id = f'well-list-id-{self.unique_id}'
        self.real_list_id = f'real-list-id-{self.unique_id}'
        self.surf_list_id = f'surf-list-id-{self.unique_id}'
        self.table_id = f'table-id-{self.unique_id}'
        self.well_tvd_id = f'well-tvd-id-{self.unique_id}'
        self.zoom_state_id = f'ui-state-id-{self.unique_id}'
        self.intersection_id = f'intersection-id-{self.unique_id}'
        self.set_callbacks(app)

    @property
    def realizations(self):
        d = {self.ensemble._realizations[real]._origpath: real
             for real in self.ensemble._realizations}
        return OrderedDict(sorted(d.items()))

    @property
    def well_names(self):
        well_folder = f'{self.well_path}/*{self.well_suffix}'
        return sorted([well for well in glob(well_folder)])

    @property
    def surface_colors(self):
        return {surf: Intersect.COLORS[i % len(Intersect.COLORS)]
                for i, surf in enumerate(self.surface_names)}

    # @cache.memoize(timeout=cache.TIMEOUT)
    # def agg_surfaces(self, surf_names, calc='avg'):
    #     agg = []
    #     for s_name in surf_names:
    #         s_arr = [self.load_surface_in_real(s_name, real).values
    #                  for real in self.ensemble._realizations]
    #         if calc=='avg':
    #             s.values = np.average(np.array(s_arr), axis=0)
    #             agg.append(s)
    #     return agg

    @cache.memoize(timeout=cache.TIMEOUT)
    def plot_xsection(self, well, reals, surf_names, tvdmin=0):
        '''Plots all lines in intersection'''
        traces = []
        for s_name in surf_names:
            traces.extend(
                make_surface_traces(
                    well, reals, s_name, self.surface_cat,
                    self.surface_colors[s_name]).to_dict('rows'))
        traces.append(make_well_trace(well, tvdmin))
        return traces

    @property
    def graph_layout(self):
        '''Styling the graph'''
        return {
            'yaxis': {'autorange': 'reversed',
                      'zeroline': False, 'title': 'TVD'},
            'xaxis': {'zeroline': False,
                      'title': 'Well horizontal distance'}

        }

    @property
    def layout(self):
        return html.Div([
            html.Div(style=Intersect.LAYOUT_STYLE, children=[
                html.Div([
                    html.P('Well:', style={'font-weight': 'bold'}),
                    dcc.Dropdown(
                        id=self.well_list_id,
                        options=[{'label': PurePath(well).stem, 'value': well}
                                 for well in self.well_names],
                        value=self.well_names[0],
                        clearable=False
                    ),
                    html.P('Realization:', style={'font-weight': 'bold'}),
                    dcc.Dropdown(
                        id=self.real_list_id,
                        options=[{'label': real, 'value': path}
                                 for path, real in self.realizations.items()],
                        value=list(self.realizations.keys())[0],
                        multi=True,
                        placeholder='All realizations'
                    ),

                    html.P('Surfaces:', style={'font-weight': 'bold'}),
                    dcc.Dropdown(
                        id=self.surf_list_id,
                        options=[{'label': r, 'value': r}
                                 for r in self.surface_names],
                        value=self.surface_names[0],
                        multi=True,
                        placeholder='All surfaces'
                    ),
                    html.Div(style=Intersect.TABLE_STYLE, children=[
                        DataTable(
                            id=self.table_id,
                            columns=[{"name": i, "id": i}
                                     for i in ['Name', 'TVDmin', 'TVDmean',
                                               'TVDmax', 'TVDstddev']]
                        )
                    ])
                ]),
                html.Div(style={'height': '80vh'}, children=[
                    dcc.Graph(style={'height': '80vh'},
                              id=self.intersection_id),
                    html.Div(style=Intersect.FENCE_OPTION_STYLE, children=[
                        html.P('Start depth:', style={'font-weight': 'bold'}),
                        html.P('Graph zoom level:', style={
                               'font-weight': 'bold'}),
                        dcc.Input(style={'overflow': 'auto'},
                                  id=self.well_tvd_id, type='number', value=0),
                        dcc.RadioItems(
                            id=self.zoom_state_id,
                            options=[{'label': k, 'value': v}
                                     for k, v in {
                                     'Keep on new data': True,
                                     'Reset on new data': False}.items()],
                            value=True,
                        ),
                    ])

                ])
            ])
        ])

    def set_callbacks(self, app):
        @app.callback(
            Output(self.intersection_id, 'figure'),
            [Input(self.well_list_id, 'value'),
             Input(self.real_list_id, 'value'),
             Input(self.surf_list_id, 'value'),
             Input(self.well_tvd_id, 'value'),
             Input(self.zoom_state_id, 'value')])
        def set_fence(_well_path, _reals, _surfs, _tvdmin, _keep_zoom_state):
            '''Callback to update intersection on data change'''
            if not isinstance(_surfs, list):
                _surfs = [_surfs]
            if not isinstance(_reals, list):
                _reals = [_reals]
            if not _reals:
                _reals = list(self.realizations.keys())
            if not _surfs:
                _surfs = self.surface_names
            s_names = [s for s in self.surface_names if s in _surfs]
            xsect = self.plot_xsection(_well_path, _reals, s_names, _tvdmin)
            layout = self.graph_layout
            if _keep_zoom_state:
                layout['uirevision'] = 'keep'
            return {'data': xsect, 'layout': layout}

        @app.callback(Output(self.table_id, 'data'),
                      [Input(self.intersection_id, 'hoverData')],
                      [State(self.intersection_id, 'figure'),
                       State(self.surf_list_id, 'value')])
        def hover(_data, _fig, _surfaces):
            '''Callback to update table on mouse over'''
            try:
                graph = _fig['data']
            except TypeError:
                return [{
                    'TVDmin': None,
                    'TVDmean': None,
                    'TVDmax': None,
                    'TVDstddev': None,
                    'Name': None
                }]
            if not isinstance(_surfaces, list):
                _surfaces = [_surfaces]
            if not _surfaces:
                _surfaces = self.surface_names
            names = {s: {'vals': [], 'min': None, 'max': None}
                     for s in _surfaces}

            for i, p in enumerate(_data['points']):
                try:
                    s_name = graph[i]['name']
                    real = self.realizations[graph[i]['real']]
                except KeyError:
                    continue
                names[s_name]['vals'].append(p['y'])
                if not names[s_name]['min']:
                    names[s_name]['min'] = p['y']
                    names[s_name]['min_real'] = real
                    names[s_name]['max'] = p['y']
                    names[s_name]['max_real'] = real
                else:
                    if names[s_name]['min'] > p['y']:
                        names[s_name]['min'] = p['y']
                        names[s_name]['min_real'] = real
                    if names[s_name]['max'] < p['y']:
                        names[s_name]['max'] = p['y']
                        names[s_name]['max_real'] = real

            return [{
                'TVDmin': f'{np.min(val["vals"]):.2f}({val["min_real"]})',
                'TVDmean': f'{np.mean(val["vals"]):.2f}',
                'TVDmax': f'{np.max(val["vals"]):.2f}({val["max_real"]})',
                'TVDstddev': f'{np.std(val["vals"]):.2f}',
                'Name': name}
                for name, val in names.items()]


@cache.memoize(timeout=cache.TIMEOUT)
def make_well_trace(well, tvdmin=0):
    '''Creates well trace for graph'''
    x = [trace[3] for trace in get_wfence(well, extend=2).values]
    y = [trace[2] for trace in get_wfence(well, extend=2).values]
    # Filter out elements less than tvdmin
    # https://stackoverflow.com/questions/17995302/filtering-two-lists-simultaneously
    try:
        y, x = zip(*((y_el, x) for y_el, x in zip(y, x) if y_el >= tvdmin))
    except ValueError:
        pass
    return {
        'x': x,
        'y': y,
        'name': PurePath(well).stem,
        'fill': None,
        'marker': {'color': 'black'}
    }


@cache.memoize(timeout=cache.TIMEOUT)
def make_surface_traces(well, reals, surf_name, cat, color):
    '''Creates surface traces for graph'''
    plot_data = []
    x = [trace[3] for trace in get_wfence(well).values]
    for j, real in enumerate(reals):
        try:
            surf = load_surface(surf_name, real, cat)
        except IOError:
            continue
        showlegend = True if j == 0 else False
        plot_data.append(
            {
                'x': x,
                'y': get_hfence(well, surf)[:, 2].copy().tolist(),
                'name': surf_name,
                'hoverinfo': 'none',
                'legendgroup': surf_name,
                'showlegend': showlegend,
                'real': real,
                'marker': {'color': color}
            })
    return pd.DataFrame(plot_data)
